# /usr/lib/python2.7
# eBASCO.py - Code to remove the baseline of accelerometric waveforms by means of tri-linear detrending of velocity time series
# Created: 5th June 2019 
# Authors: Maria D'Amico, Chiara Felicetta and Erica Schiappapietra
# version 1.0

# License
# --------
# eBASCO.py code is licensed under a Creative Commons Attribution-NoDerivatives 4.0 International License (cfr. Section 4 Sui Generis Database Rights)

import datetime
import time
import pyasdf # (pyasdf==0.4.0)
import argparse
from os.path import isfile as os_path_isfile
from obspy.core import trace, Stream # (obspy==1.1.1)
from obspy.signal.filter import lowpass # (obspy==1.1.1)
from obspy.signal.invsim import cosine_taper as cosTaper # (obspy==1.1.1)
from copy import deepcopy as copy_deepcopy
from numpy import arange, logspace, cumsum, gradient, zeros, abs,log10, array, concatenate, argmin, argmax, asarray, ones # (numpy==1.16.2)

from scipy import optimize, stats # (scipy==1.2.1)
from os import getcwd as os_getcwd 
from os import environ as os_environ
from os.path import basename as os_path_basename 
from sys import argv as sys_argv
from sys import exit as sys_exit
from math import isnan
from matplotlib.mlab import detrend_linear # (matplotlib==2.2.4)
import matplotlib as mpl
if os_environ.get('DISPLAY','') == '':
    print ('no disply found. Using non-interactive Agg backend')
    mpl.use('Agg')
from matplotlib import pyplot as plt 
import psutil #(psutil==5.6.2)
import resp_spectra_f64
import traceback

# from cachetools import cached, TTLCache
# cache = TTLCache(maxsize=100, ttl=300)    
# @cached(cache)

def eBASCO(opts_str):

    start = time.time()
    stop_time = start + 20	
    # command line HELP
    synopsis = "To Process waveforms in asdf file with eBASCO method"

    p = argparse.ArgumentParser(description=synopsis)         
    
    p.add_argument("--ifile", "--input-file", action="store", dest="ifile", type=str, help="ASDF file containing the waveforms to be processed")
    p.add_argument("--ofldr", "--output-folder", action="store", dest="ofldr", type=str, help="Folder containing the ASDF file with eBASCO processed waveforms")
    
    p.add_argument("--t1", "--t1-samples", action="store", dest="t1", type=int, default=5, help="number of T1 correction points samples") 
    p.add_argument("--t2", "--t2-samples", action="store", dest="t2", type=int, default=20, help="number of T2 correction points samples")        
    p.add_argument("--t3", "--t3-samples", action="store", dest="t3", type=int, default=20, help="number of T3 correction points samples")    
    p.add_argument("--eps", "--epsilon", action="store", dest="eps", type=float, default=0.25, help="maximum threshold of the relative difference between acceleration in T1 and T2 correction point before and after the eBASCO processing")    
    p.add_argument("--mfst", "--mf-start", action="store", dest="mfst", type=float, default=1.5, help="multiplier factor to cut the beginning of the waveform on the base of the energy criterion")    
    p.add_argument("--mfnd", "--mf-end", action="store", dest="mfnd", type=float, default=2.0, help="multiplier factor to cut the end of the waveform on the base of the energy criterion")    
    p.add_argument("--ca", "--cut-a", action="store", dest="ca", type=float, default=0, help="seconds to cut from the beginning of the waveforms; ca is used only if different from the default value, otherwise eBASCO use the energy criterion to cut the strong-phase of motion between the 5% and the 95% of the energy release")
    p.add_argument("--cz", "--cut-z", action="store", dest="cz", type=float, default=0, help="seconds to cut from the end of the waveforms; cz is used only if different from the default value, otherwise eBASCO use the energy criterion to cut the strong-phase of motion between the 5% and the 95% of the energy release")
    p.add_argument("--ta", "--taper-a", action="store", dest="ta", type=float, default=5, help="percentage of the signal length to set the time window at the beginning of the waveform where a cosine taper will be applied")	
    p.add_argument("--he", "--high-cutoff-freq-ew", action="store", dest="he", type=float, default=35, help="cutoff frequency of the low-pass Butterworth filter (first component)")
    p.add_argument("--hn", "--high-cutoff-freq-ns", action="store", dest="hn", type=float, default=35, help="cutoff frequency of the low-pass Butterworth filter (second component)")
    p.add_argument("--hz", "--high-cutoff-freq-up", action="store", dest="hz", type=float, default=35, help="cutoff frequency of the low-pass Butterworth filter (third component)")	
    p.add_argument("--fo", "--filter-order", action="store", dest="fo", type=float, default=2, help="Butterworth filter order")
    p.add_argument("--pro", "--proc-typ", action="store", dest="pr", type=str, default='manual2', help="Type of Processing (default = 'manual2')")

    p.add_argument("--v", "--verbose", action="store", dest="verbose")

    if len(opts_str.split(' ')) < 2 : p.error("incorrect number of arguments, the input file [--ifile IFILE] and the output folder [ofldr OFOLDER] are mandatory"); p.print_help()
        
    opts = p.parse_args(opts_str.split(' '))
    
    orig_dirname = os_getcwd()
    orig_pyversion = 'python'
    orig_processing_code = 'eBASCO.py'
    orig_filename = os_path_basename(opts.ifile)
    out_name = opts.ofldr + opts.ifile.split('/')[-1] + ".mp_mb_processed"
    out_png =  out_name + '.png'

    valid_asdf(opts.ifile)
	
	# define the command line	
    commandLine = orig_pyversion + ' ' + orig_processing_code + ' ' + "--ifile=" + str(opts.ifile)+ ' ' + "--ofldr=" + str(opts.ofldr) + ' ' + "--t1=" + str(opts.t1) + ' ' + "--t2="+ str(opts.t2) + ' ' + "--t3=" + str(opts.t3) + ' ' + "--eps=" + str(opts.eps) + ' ' + "--mfst=" + str(opts.mfst) + ' ' + "--mfnd=" + str(opts.mfnd) + ' ' +"--ca=" + str(opts.ca) + ' ' + "--cz=" + str(opts.cz) + ' ' + "--ta="+ str(opts.ta) + ' ' + "--he=" + str(opts.he) + ' ' + "--hn=" + str(opts.hn) + ' ' + "--hz=" + str(opts.hz) + ' ' + "--fo=" + str(opts.fo) 
				
	# read ASDF format
    ds = pyasdf.ASDFDataSet(opts.ifile, mode='r')
    
    # check ASDF input file
    cnt = 0
    for station in ds.waveforms:
        for tag in sorted(station.get_waveform_tags()):
            if tag[-6:] == "acc_cv":
                cnt += 1
                if cnt == 1:
                    st_e = station[tag]
                    head_e = ds.auxiliary_data.Headers[st_e[0].stats.network + '_' + st_e[0].stats.station][tag].parameters
                    tag_e = tag
                if cnt == 2:
                    st_n = station[tag]
                    head_n = ds.auxiliary_data.Headers[st_n[0].stats.network + '_' + st_n[0].stats.station][tag].parameters
                    tag_n = tag
                if cnt == 3:
                    st_z = station[tag]
                    head_z = ds.auxiliary_data.Headers[st_z[0].stats.network + '_' + st_z[0].stats.station][tag].parameters
                    tag_z = tag
                if opts.verbose: print (tag) 

    if cnt != 3:
        err_msg = "ERROR: the number of input '*_cv' streams is not equal to three"
        print(err_msg)
        return dict_fmt(exit_status=2, exit_message=err_msg)    
        
    if (st_e[0].stats.channel[-1] == 'E' and st_n[0].stats.channel[-1] == 'N' and st_z[0].stats.channel[-1] == 'Z'): pass
    elif (st_e[0].stats.channel[-1] == '1' and st_n[0].stats.channel[-1] == '2' and st_z[0].stats.channel[-1] == 'Z'): pass
    elif (st_e[0].stats.channel[-1] == '2' and st_n[0].stats.channel[-1] == '3' and st_z[0].stats.channel[-1] == 'Z'): pass
    elif (st_e[0].stats.channel[-1] == 'A' and st_n[0].stats.channel[-1] == 'B' and st_z[0].stats.channel[-1] == 'C'): pass
    elif (st_e[0].stats.channel[-1] == 'U' and st_n[0].stats.channel[-1] == 'V' and st_z[0].stats.channel[-1] == 'W'): pass
    elif (st_e[0].stats.channel[-1] == 'R' and st_n[0].stats.channel[-1] == 'T' and st_z[0].stats.channel[-1] == 'Z'): pass
    else:
        err_msg = "ERROR: WRONG ORDER FOR INPUT FILE (RIGHT ORDER: E,N,Z)"
        print(err_msg)
        return dict_fmt(exit_status=2, exit_message=err_msg) 

    if opts.t1 > 0: pass
    else:
        err_msg = "ERROR: THE NUMBER OF THE T1 SAMPLES MUST BE LARGER THAN 0"
        return dict_fmt(exit_status=2, exit_message=err_msg)

    if opts.t2 > 0: pass 
    else:
        err_msg = "ERROR: THE NUMBER OF THE T2 SAMPLES MUST BE LARGER THAN 0"
        return dict_fmt(exit_status=2, exit_message=err_msg)

    if opts.t3 > 0: pass 
    else:
        err_msg = "ERROR: THE NUMBER OF THE T3 SAMPLES MUST BE LARGER THAN 0"
        return dict_fmt(exit_status=2, exit_message=err_msg)
        

    if opts.eps>=0 and opts.eps<=1: pass
    else:
        err_msg = "ERROR: WRONG VALUE OF THE TRESHOLD USED TO CHECK THE GOODNESS OF THE ACCELERATION TIME HISTORIES,  THE TRESHOLD MUST BE WITHIN THE INTERVAL [0,1] (es: 0.10, 0.25)"
        return dict_fmt(exit_status=2, exit_message=err_msg)
        
    if opts.mfst > 0: pass
    else:
        err_msg = "ERROR: WRONG VALUE OF THE MULTIPLIER FACTOR USED TO CUT THE BEGINNIG OF THE SIGNAL BY USING THE ENERGY CRITERION, MULTIPLIER FACTOR MUST BE LARGER THAN 0 (es: 1.5, 2.2)"
        return dict_fmt(exit_status=2, exit_message=err_msg)                

    if opts.mfnd > 0: pass
    else:
        err_msg = "ERROR: WRONG VALUE OF THE THE MULTIPLIER FACTOR USED TO CUT THE END OF THE SIGNAL BY USING THE ENERGY CRITERION, MULTIPLIER FACTOR MUST BE LARGER THAN 0 (es: 1.5, 2.2)"
        print(err_msg)
        return dict_fmt(exit_status=2, exit_message=err_msg) 


    if opts.cz >= 0: pass
    else:
        err_msg = "ERROR: WRONG VALUE OF THE SECONDS TO CUT AT THE END OF THE SIGNAL, SECOND TO CUT MUST BE LARGER THAN 0 (es: 5.0, 20, 25.45)"
        return dict_fmt(exit_status=2, exit_message=err_msg)
                                
    if isnan(opts.ta)== False \
    or isinstance(opts.ta,int) or isinstance(opts.ta,float): pass 
    else:
        err_msg = "ERROR: MISSING VALUE OR WRONG FORMAT FOR THE TAPER PERCENTAGE AT THE BEGINNING OF THE SIGNAL (RIGHT FORMAT: 'float' or 'int' - es: 5.0, 20, 25.45)"
        print(err_msg)
        return dict_fmt(exit_status=2, exit_message=err_msg)

    if (opts.ta < 0 or opts.ta > 50):
        err_msg = ("ERROR: DEPRECATED VALUE FOR THE TAPER PERCENTAGE AT THE BEGINNING OF THE SIGNAL; IT MUST BE WITHIN THE INTERVAL ]0,50[ ")    
        print(err_msg)
        return dict_fmt(exit_status=2, exit_message=err_msg)
        
    # read the station list in the ASDF file
    STA_LIST = ds.waveforms.list()
    curr_sta = STA_LIST[0]

    wf_tags = ds.waveforms[curr_sta].get_waveform_tags()


    print ("#########################################")   
    print ("")        
    print ("working station: " + curr_sta)
    print ("")
    print ("input file: " + orig_dirname + "/" + opts.ifile)
    print ("output file: " + orig_dirname + "/" + out_name)  
    print ("")
    print ("#########################################")
    print ("")
    

                
    # read raw and corrected waveforms and related metadata in the ASDF file 
    wf_md_E = wf_md_Extract(ds,curr_sta,tag_e,tag_e.replace('_acc_cv','_acc_mp'))
    wf_md_N = wf_md_Extract(ds,curr_sta,tag_n,tag_n.replace('_acc_cv','_acc_mp'))
    wf_md_Z = wf_md_Extract(ds,curr_sta,tag_z,tag_z.replace('_acc_cv','_acc_mp'))      

    # check if the instrument is Digital
    if ds.auxiliary_data.Headers[curr_sta.replace('.','_')][tag_e].parameters['instrument_analog_digital'] == b'A':
        print("WARNING: analogic waveforms!!!")
        
    # check if the Analog waveform is LT (Late Triggered)
    if ds.auxiliary_data.Headers[curr_sta.replace('.','_')][tag_e].parameters['late_normal_triggered'] == b'LT': 
        print("WARNING: Late Triggered data")
        
    # create temporary arrays 
    strms = dpcpy_Stream(ds.waveforms[curr_sta][tag_e], ds.waveforms[curr_sta][tag_n], ds.waveforms[curr_sta][tag_z]) # make a copy of streams

    dur_e = strms[0][0].stats.endtime - strms[0][0].stats.starttime
    dur_n = strms[1][0].stats.endtime - strms[1][0].stats.starttime
    dur_z = strms[2][0].stats.endtime - strms[2][0].stats.starttime
    dur_min = min(dur_e, dur_n, dur_z)
    npts_max = max(strms[0][0].stats.npts, strms[1][0].stats.npts, strms[2][0].stats.npts)
    if (opts.ca + opts.cz) > dur_min: 
        p.error("cannot cut " + str(opts.ca + opts.cz) + " s from a " + str(dur_min) + " s signal!!!")
        
    #----------------               	                    
    # CUT ALL STREAMS
    #----------------
    print ("")
    print ("----------CUT ALL STREAMS------------")
    print("")
    print ("DATE_TIME_FIRST_SAMPLE_YYYYMMDD_HHMMSS.000: %s" % fromUTCDateTime(max(strms[0][0].stats.starttime, \
    strms[1][0].stats.starttime, strms[2][0].stats.starttime)))
    print ("DATE_TIME_LAST_SAMPLE_YYYYMMDD_HHMMSS.000: %s" % fromUTCDateTime(max(strms[0][0].stats.endtime, \
    strms[1][0].stats.endtime, strms[2][0].stats.endtime))) 
    print("")
    print("--------raw duration---------") 
    print("")
    print("dur EW: ", strms[0][0].stats.endtime-strms[0][0].stats.starttime)
    print("dur NS: ", strms[1][0].stats.endtime-strms[1][0].stats.starttime)
    print("dur UD: ", strms[2][0].stats.endtime-strms[2][0].stats.starttime)


    if opts.ca == 0 and opts.cz == 0:
        print("")
        print ("--> cut all streams according to energy criterion")
        print("")
    else:
        print("")
        print ("--> cut all streams to the same length, according to max_starttime / min_endtime and user settings (ca / cz) ")
        print("")	 
    

    strms = dpcpy_Stream(ds.waveforms[curr_sta][tag_e], ds.waveforms[curr_sta][tag_n], ds.waveforms[curr_sta][tag_z]) # make a copy of streams
    
    # cut all streams according to energy criterion
    [ACC_E_strm_1, ACC_N_strm_1, ACC_Z_strm_1, cut_start, cut_end] = cut_wf_ARIAS (strms[0],strms[1],strms[2], wf_md_E["sampling_rate"], wf_md_N["sampling_rate"], wf_md_Z["sampling_rate"],0.05,0.95, opts.ca, opts.cz,opts.mfst,opts.mfnd)
                             
    # cut all streams to the same length, according to max_starttime / min_endtime and user settings (ca / cz) 
    
    [ACC_E_strm, ACC_N_strm, ACC_Z_strm] = wf_Slice(ACC_E_strm_1, ACC_N_strm_1, ACC_Z_strm_1, opts.ca, opts.cz)            
    
    WF_CUT = ACCtoDISP (ACC_E_strm, ACC_N_strm, ACC_Z_strm, wf_md_E["sampling_rate"], wf_md_N["sampling_rate"], wf_md_Z["sampling_rate"])
    print("")
    print("-----------updated duration---------") 
    print("")
    print("dur EW: ", ACC_E_strm[0].stats.endtime-ACC_E_strm[0].stats.starttime)
    print("dur NS: ", ACC_N_strm[0].stats.endtime-ACC_N_strm[0].stats.starttime)
    print("dur UD: ", ACC_Z_strm[0].stats.endtime-ACC_Z_strm[0].stats.starttime)
    print("")


    START_CUT_E = ACC_E_strm[0].stats.starttime - strms[0][0].stats.starttime 
    START_CUT_N = ACC_N_strm[0].stats.starttime - strms[1][0].stats.starttime
    START_CUT_Z = ACC_Z_strm[0].stats.starttime - strms[2][0].stats.starttime

    END_CUT_E = strms[0][0].stats.endtime - ACC_E_strm[0].stats.endtime
    END_CUT_N = strms[1][0].stats.endtime - ACC_N_strm[0].stats.endtime  
    END_CUT_Z = strms[2][0].stats.endtime - ACC_Z_strm[0].stats.endtime

    
    if ACC_E_strm[0].stats.endtime-ACC_E_strm[0].stats.starttime == 0 or ACC_N_strm[0].stats.endtime-ACC_N_strm[0].stats.starttime == 0 or ACC_Z_strm[0].stats.endtime-ACC_Z_strm[0].stats.starttime == 0:
        err_msg = "WARNING!! no samples in the cutted data"
        print(err_msg)
        return dict_fmt(exit_status=2, exit_message=err_msg)
    start_time = ACC_E_strm[0].stats.starttime        
    end_time = ACC_E_strm[0].stats.endtime
    
    #------------------------------------------------
    # TRILINEAR DETREND OF THE VELOCITY TIME HISTORY
    #------------------------------------------------
    
    print ("")
    print ("----------TRILINEAR DETREND OF THE VELOCITY TIME HISTORY------------")
    print("")
    print("--> sample the T1 time before the 5% of the cumulated energy (ARIAS INTENSITY)")
    print("")  
    # sample the T1 time before the 5% of the cumulated energy (ARIAS INTENSITY)
    
    T1_SAMPLES = sample_T1 (WF_CUT[0][0],WF_CUT[0][1],WF_CUT[0][2],wf_md_E["sampling_rate"],wf_md_N["sampling_rate"],wf_md_Z["sampling_rate"],opts.t3)

    print("--> linear de-trend between the first sample and T1  (velocity)")
    print("")
    # define the "pre-event" between the first sample T0 and T1  (velocity)
    [PRE_EVE_TIME, PRE_EVE_VEL] = pre_eve(WF_CUT[1][0],WF_CUT[1][1],WF_CUT[1][2],wf_md_E["sampling_rate"],wf_md_N["sampling_rate"],wf_md_Z["sampling_rate"],T1_SAMPLES)  
    # linear de-trend in the "pre-event"
    try:
        PRE_EVE_VEL_DET, LAST_VALUE_LIN_FIT, LIN_FIT = pre_eve_detrend (PRE_EVE_TIME,PRE_EVE_VEL)
    except:
            err_msg = "ERROR! bad cut of the pre-event, please try it again!"
            return dict_fmt(exit_status=2, exit_message=err_msg)        
    print("--> sample T3 time after the 5% of the cumulated energy (ARIAS INTENSITY)")
    print("")
    # sample T3 time after the 5% of the cumulated energy (ARIAS INTENSITY)
    T3_SAMPLES = sample_T3 (WF_CUT[0][0],WF_CUT[0][1],WF_CUT[0][2],wf_md_E["sampling_rate"],wf_md_N["sampling_rate"],wf_md_Z["sampling_rate"],opts.t3)
    
    # search the end points for each component
    END_POINTS = end_point (WF_CUT[0][0],WF_CUT[0][1],WF_CUT[0][2],wf_md_E["sampling_rate"],wf_md_N["sampling_rate"],wf_md_Z["sampling_rate"])
    
    print("--> sample T2 time from each T3 and the end of the signal and detrend the velocity trace in the time windows start-T1, T1-T2, T3-Tend ")
    print("")
    T1_E = []; T2_E = []; T3_E = []; TIME_E=[]; VEL_CORR_E=[]
    T1_N = []; T2_N = []; T3_N = []; TIME_N=[]; VEL_CORR_N=[]
    T1_Z = []; T2_Z = []; T3_Z = []; TIME_Z=[]; VEL_CORR_Z=[]
    
    # memory_usage = psutil.virtual_memory()
    # while time.time() < stop_time:
    # while memory_usage.percent < 75:
    for index_T1 in arange(len(T1_SAMPLES[0])-int(len(T1_SAMPLES[0])/2)):
        memory_usage = psutil.virtual_memory()
        #print(memory_usage.percent) 
        if memory_usage.percent < 50: 
            for index_T3 in arange(len(T3_SAMPLES[0])-int(len(T3_SAMPLES[0])/2)):
                                                  
                # sample T2 time from each T3 and the end of the signal
                T2_SAMPLES = T2_sample(T3_SAMPLES,END_POINTS,index_T3, opts.t2)
                            
                for index_T2 in arange(len(T2_SAMPLES[0])-int(len(T2_SAMPLES[0])/2)): 
                                
                    # define the "post event" between T2 and the end of the signal (velocity)
                    [POST_EVE_TIME, POST_EVE_VEL] = post_eve(WF_CUT[1][0],WF_CUT[1][1],WF_CUT[1][2],wf_md_E["sampling_rate"],wf_md_N["sampling_rate"],wf_md_Z["sampling_rate"],T2_SAMPLES,END_POINTS,index_T2)  
                    # linear de-trend between T2 and the end of the signal (velocity)
                    [POST_EVE_VEL_DET, POST_EVE_LIN_FIT] = post_eve_detrend (POST_EVE_TIME,POST_EVE_VEL)
                    
                    # define the "strong-event" between T1 and T2 (velocity)                   
                    [STRONG_PHASE_TIME, STRONG_PHASE_VEL] = strong_eve(WF_CUT[1][0],WF_CUT[1][1],WF_CUT[1][2],wf_md_E["sampling_rate"],wf_md_N["sampling_rate"],wf_md_Z["sampling_rate"],T1_SAMPLES,T2_SAMPLES,index_T1,index_T2)
                    
                    # calculate the slope Am of the line between V(T1) and V(T2); V(T1) and V(T2) are the last point of the detrending line in the "pre-event" and the first point of the detrending line in the "post-event", respectively
                    Am_E = float((POST_EVE_LIN_FIT[0][0]-LAST_VALUE_LIN_FIT[0][index_T1])/(T2_SAMPLES[0][index_T2]-T1_SAMPLES[0][index_T1]))
                    Am_N = float((POST_EVE_LIN_FIT[1][0]-LAST_VALUE_LIN_FIT[1][index_T1])/(T2_SAMPLES[1][index_T2]-T1_SAMPLES[1][index_T1]))
                    Am_Z = float((POST_EVE_LIN_FIT[2][0]-LAST_VALUE_LIN_FIT[2][index_T1])/(T2_SAMPLES[2][index_T2]-T1_SAMPLES[2][index_T1]))
                    # calculate the ordinate intercept of the line between V(T1) and V(T2)
                    q_E = LAST_VALUE_LIN_FIT[0][index_T1] - Am_E*T1_SAMPLES[0][index_T1]
                    q_N = LAST_VALUE_LIN_FIT[1][index_T1] - Am_N*T1_SAMPLES[1][index_T1]
                    q_Z = LAST_VALUE_LIN_FIT[2][index_T1] - Am_Z*T1_SAMPLES[2][index_T1]
                    
                    def retta(x,m,q):
                            return m*x + q
                    # define the detrending line between V(T1) and V(T2)                        
                    STRONG_PHASE_LINE = [retta(STRONG_PHASE_TIME[0],Am_E,q_E), retta(STRONG_PHASE_TIME[1],Am_N,q_N), retta(STRONG_PHASE_TIME[2],Am_Z,q_Z) ]
                    # linear de-trend in the "strong-event" 
                    STRONG_PHASE_VEL_DET = strong_eve_detrend (STRONG_PHASE_TIME,STRONG_PHASE_VEL,STRONG_PHASE_LINE)
		            
		            # concatenate the trilinear detrended velocity trace (time windows start-T1, T1-T2, T3-Tend)
                    TIME_E.append(concatenate((PRE_EVE_TIME[0][index_T1],STRONG_PHASE_TIME[0],POST_EVE_TIME[0][1:]),axis=0))
                    VEL_CORR_E.append(concatenate((PRE_EVE_VEL_DET[0][index_T1],STRONG_PHASE_VEL_DET[0],POST_EVE_VEL_DET[0][1:]),axis=0))                                           
                    TIME_N.append(concatenate((PRE_EVE_TIME[1][index_T1],STRONG_PHASE_TIME[1],POST_EVE_TIME[1][1:]),axis=0))
                    VEL_CORR_N.append(concatenate((PRE_EVE_VEL_DET[1][index_T1],STRONG_PHASE_VEL_DET[1],POST_EVE_VEL_DET[1][1:]),axis=0)) 
                     
                    TIME_Z.append(concatenate((PRE_EVE_TIME[2][index_T1],STRONG_PHASE_TIME[2],POST_EVE_TIME[2][1:]),axis=0))
                    VEL_CORR_Z.append(concatenate((PRE_EVE_VEL_DET[2][index_T1],STRONG_PHASE_VEL_DET[2],POST_EVE_VEL_DET[2][1:]),axis=0))    
                    
                    T1_E.append(T1_SAMPLES[0][index_T1]); T1_N.append(T1_SAMPLES[1][index_T1]); T1_Z.append(T1_SAMPLES[2][index_T1])                 
                    T2_E.append(T2_SAMPLES[0][index_T2]); T2_N.append(T2_SAMPLES[1][index_T2]); T2_Z.append(T2_SAMPLES[2][index_T2])
                    T3_E.append(T3_SAMPLES[0][index_T3]); T3_N.append(T3_SAMPLES[1][index_T3]); T3_Z.append(T3_SAMPLES[2][index_T3])              
        else: 
            print("       ")
            print("#------------------# WARNING!!!! very high percentage memory usage or trace too long, killed processing #------------------#")
            print("       ")
            return dict_fmt(exit_status=0, exit_message="percentage of memory usage too high, killed processing", in_file=opts.ifile, out_file=ds_mp_mb.filename, out_format='hdf5')

    #---------------------------------------
    # DIFFERENTIATE THE VELOCITY
    #---------------------------------------
    print("--> differentiate the corrected velocity to obtaine the acceleration")   
    print("")             
    # differentiate the corrected velocity to obtaine the acceleration
    strms = dpcpy_Stream(VEL_CORR_E, VEL_CORR_N, VEL_CORR_Z) # make a copy of the velocity stream for each component
    ACC_CORR_E = []; ACC_CORR_N = []; ACC_CORR_Z = []
    for ind in arange(len(strms[0])):
        ACC_CORR_E.append(gradient(strms[0][ind],wf_md_E["sampling_rate"]))
    for ind in arange(len(strms[1])):
        ACC_CORR_N.append(gradient(strms[1][ind],wf_md_N["sampling_rate"]))
    for ind in arange(len(strms[2])):
        ACC_CORR_Z.append(gradient(strms[2][ind],wf_md_Z["sampling_rate"]))
        
    #--------------------------------------------
    # CHECK THE OPTS.T1*OPTS.T2*OPTS.T3 SOLUTIONS
    #--------------------------------------------
    
    print("--> check what solutions are acceptable")
    print("")
    # check what solutions satisfy the treshold between ACC_RAW and ACC_eBASCO in T1 and T2; the default treshold corresponds to the 25% of the difference between ACC_RAW and ACC_eBASCO in T1 and T2 (absolute value)
    [ACC_CORR_E_good,ACC_bad_E,T3_E_good,INDEX_good_E] = acc_corr(WF_CUT[0][0],ACC_CORR_E,wf_md_E["sampling_rate"],T1_E,T2_E,T3_E,opts.eps)             
    [ACC_CORR_N_good,ACC_bad_N,T3_N_good,INDEX_good_N] = acc_corr(WF_CUT[0][1],ACC_CORR_N,wf_md_N["sampling_rate"],T1_N,T2_N,T3_N,opts.eps)
    [ACC_CORR_Z_good,ACC_bad_Z,T3_Z_good,INDEX_good_Z] = acc_corr(WF_CUT[0][2],ACC_CORR_Z,wf_md_Z["sampling_rate"],T1_Z,T2_Z,T3_Z,opts.eps)             
    
    print("--> calculate VEL and DISP for each acceptable ACC") 
    print("")           
    #  calculate VEL and DISP for each acceptable ACC
    [TIME_E_good, VEL_CORR_E_good, DISP_CORR_E_good] = accept_solution(ACC_CORR_E_good, wf_md_E["sampling_rate"], 'EW')
    [TIME_N_good, VEL_CORR_N_good, DISP_CORR_N_good] = accept_solution(ACC_CORR_N_good, wf_md_N["sampling_rate"], 'NS')
    [TIME_Z_good, VEL_CORR_Z_good, DISP_CORR_Z_good] = accept_solution(ACC_CORR_Z_good, wf_md_Z["sampling_rate"], 'UD')
            
   
    # to calculate the flatness indicator
    # ----------------------------------#
    # r = linear correlation coefficient of the samples between T3 and the end of the signal --> -1<= r <=1   if |r|=1 good linear covariance
    # b = slope of the regression line between T3 and the end of the signal --> 0 when the [T3:end] signal is costant
    # sigma = variance of the mean value of the samples between T3 and the end of the signal --> sigma-->0 means that the samples are poor scattered around the mean value
    # f_value = |r|/(|b|*sigma) is the flatness indicator  f--> high value when the signal between T3 and the end of the signal approssimates a line parallel to the time axes
    #----------------------------------#
    
    print("")
    print("--> identify the best solution througth the flatness indicator")
    print("")
    print("------------FLATNESS INDICATORS--------------")       

    st_E_dis=[-999]; st_N_dis=[-999]; st_Z_dis=[-999]

    if len(ACC_CORR_E_good)>0: f_value_E = []; f_value_E = fvalue(TIME_E_good,DISP_CORR_E_good,T3_E_good); num_E = argmax(f_value_E);[st_E_acc, st_E_vel, st_E_dis] = final_trace(ACC_CORR_E_good,num_E,opts.ta,wf_md_E["sampling_rate"],opts.fo,opts.he); print("f value EW: ",format(f_value_E[num_E],'.0f')); T1_E_good = T1_E[num_E]; T2_E_good = T2_E[num_E]; T3_E_good = T3_E[num_E];
    else: num_E = -999; st_E_acc=[0]
    if len(ACC_CORR_N_good)>0: f_value_N = []; f_value_N = fvalue(TIME_N_good,DISP_CORR_N_good,T3_N_good); num_N = argmax(f_value_N);[st_N_acc, st_N_vel, st_N_dis] = final_trace(ACC_CORR_N_good,num_N,opts.ta,wf_md_N["sampling_rate"],opts.fo,opts.hn); print("f value NS: ",format(f_value_N[num_N],'.0f')); T1_N_good = T1_N[num_N]; T2_N_good = T2_N[num_N]; T3_N_good = T3_N[num_N];
    else: num_N = -999; st_N_acc=[0]
    if len(ACC_CORR_Z_good)>0: f_value_Z = []; f_value_Z = fvalue(TIME_Z_good,DISP_CORR_Z_good,T3_Z_good); num_Z = argmax(f_value_Z);[st_Z_acc, st_Z_vel, st_Z_dis] = final_trace(ACC_CORR_Z_good,num_Z,opts.ta,wf_md_Z["sampling_rate"],opts.fo,opts.hz); print("f value UD: ",format(f_value_Z[num_Z],'.0f')); T1_Z_good = T1_Z[num_Z]; T2_Z_good = T2_Z[num_Z]; T3_Z_good = T3_Z[num_Z];
    else: num_Z = -999; st_Z_acc=[0]
    print("---------------------------------------------") 
    print("")

    if (num_E == -999 and num_N==-999) and num_Z==-999:
        err_msg = "there are no acceptable solutions, please try to change default options!"
        return dict_fmt(exit_status=2, exit_message=err_msg)     
  
    #--------------------------------------------
    # READ THE PERMANET DISPLACEMENT
    #--------------------------------------------            
    print("")
    print("------------PERMANENT DISPLACEMENT--------------") 
    EW_pd = st_E_dis[-1]; NS_pd = st_N_dis[-1]; UD_pd = st_Z_dis[-1]
    print("permanent displacement EW [cm]: ",format(EW_pd,'.2f'))
    print("permanent displacement NS [cm]: ",format(NS_pd,'.2f'))
    print("permanent displacement UD [cm]: ",format(UD_pd,'.2f'))
    print("---------------------------------------------") 
    print("")
    

    #---------------------------------------------------
    # CALCULATE DISPLACEMENT  AND ACCELERATION SPECTRA 
    #----------------------------------------------------

    periods = array([0.010, 0.020, 0.022, 0.025, 0.029, 0.030, 0.032, 0.035, 0.036, 0.040, 0.042, 0.044, 0.045, 0.046, 0.048, 0.050, 0.055, 0.060, 0.065, 0.067, 0.070, 0.075, 0.080, 0.085, 0.090, 0.095, 0.100, 0.110, 0.120, 0.130, 0.133, 0.140, 0.150, 0.160, 0.170, 0.180, 0.190, 0.200, 0.220, 0.240, 0.250, 0.260, 0.280, 0.290, 0.300, 0.320, 0.340, 0.350, 0.360, 0.380, 0.400, 0.420, 0.440, 0.450, 0.460, 0.480, 0.500, 0.550, 0.600, 0.650, 0.667, 0.700, 0.750, 0.800, 0.850, 0.900, 0.950, 1.000, 1.100, 1.200, 1.300, 1.400, 1.500, 1.600, 1.700, 1.800, 1.900, 2.000, 2.200, 2.400, 2.500, 2.600, 2.800, 3.000, 3.200, 3.400, 3.500, 3.600, 3.800, 4.000, 4.200, 4.400, 4.600, 4.800, 5.000, 5.500, 6.000, 6.500, 7.000, 7.500, 8.000, 8.500, 9.000, 9.500, 10.000])
    
    damping_val = 0.05

    
    if EW_pd != -999: [st_E_acc, st_E_vel, st_E_dis] = update_ebasco_trace(curr_sta,tag_e,ds,start_time,end_time,st_E_acc,st_E_vel,st_E_dis)
    if NS_pd != -999: [st_N_acc, st_N_vel, st_N_dis] = update_ebasco_trace(curr_sta,tag_n,ds,start_time,end_time,st_N_acc,st_N_vel,st_N_dis)
    if UD_pd != -999: [st_Z_acc, st_Z_vel, st_Z_dis] = update_ebasco_trace(curr_sta,tag_z,ds,start_time,end_time,st_Z_acc,st_Z_vel,st_Z_dis)
    
    if EW_pd != -999:xsd_E,xsa_E,std_E,sta_E = resp_spectra_f64.pcn05(st_E_acc.stats.npts,len(periods),st_E_acc.stats.delta,damping_val,periods,st_E_acc.data)
    if NS_pd != -999:xsd_N,xsa_N,std_N,sta_N = resp_spectra_f64.pcn05(st_N_acc.stats.npts,len(periods),st_N_acc.stats.delta,damping_val,periods,st_N_acc.data)
    if UD_pd != -999:xsd_Z,xsa_Z,std_Z,sta_Z = resp_spectra_f64.pcn05(st_Z_acc.stats.npts,len(periods),st_Z_acc.stats.delta,damping_val,periods,st_Z_acc.data)
    
    (st_E_sa, st_N_sa, st_Z_sa) = dpcpy_Stream(st_E_acc, st_N_acc, st_Z_acc) # make a copy of streams     
    if EW_pd != -999: st_E_sa = add_spectra(st_E_sa,periods,xsa_E)
    if NS_pd != -999: st_N_sa = add_spectra(st_N_sa,periods,xsa_N)
    if UD_pd != -999: st_Z_sa = add_spectra(st_Z_sa,periods,xsa_Z)
    
    (st_E_sd, st_N_sd, st_Z_sd) = dpcpy_Stream(st_E_dis, st_N_dis, st_Z_dis) # make a copy of streams 
    if EW_pd != -999: st_E_sd = add_spectra(st_E_sd,periods,xsd_E)
    if NS_pd != -999: st_N_sd = add_spectra(st_N_sd,periods,xsd_N)
    if UD_pd != -999: st_Z_sd = add_spectra(st_Z_sd,periods,xsd_Z)
    
    #-------------------------------------------#
    #				OUTPUT						#
    #-------------------------------------------#

    if num_E != -999:
        pre_tag_path = st_E_acc.stats.network + "_" + st_E_acc.stats.station + "/"; pass
    elif num_N != -999:
        pre_tag_path = st_N_acc.stats.network + "_" + st_N_acc.stats.station + "/"; pass
    elif num_Z != -999:
        pre_tag_path = st_Z_acc.stats.network + "_" + st_Z_acc.stats.station + "/"
        
    proc_dict = {'automatic': 'ap', 'manual1': 'mp', 'manual2': 'mb'}
    proc_type = proc_dict[opts.pr]
    
    evid = str(ds.events[0].resource_id).split('=')[1]
    out_name = st_z[0].stats.network + '.' + st_z[0].stats.station + '.' + st_z[0].stats.location + '.' + st_z[0].stats.channel[0:2] + '.' + evid
    ds_mp_mb = pyasdf.ASDFDataSet(opts.ofldr + out_name + ".h5_mp_mb_processed", mode='w', compression=None)
    
    
    # add cv waveforms
    ds_mp_mb.add_waveforms(st_e, tag=tag_e, event_id=evid)
    ds_mp_mb.add_waveforms(st_n, tag=tag_n, event_id=evid)
    ds_mp_mb.add_waveforms(st_z, tag=tag_z, event_id=evid)
    
    # add cv headers
    ds_mp_mb.add_auxiliary_data(data=zeros(0), data_type="Headers", path=pre_tag_path + tag_e, parameters=head_e)
    ds_mp_mb.add_auxiliary_data(data=zeros(0), data_type="Headers", path=pre_tag_path + tag_n, parameters=head_n)
    ds_mp_mb.add_auxiliary_data(data=zeros(0), data_type="Headers", path=pre_tag_path + tag_z, parameters=head_z)    

    # update waweforms #
    # define tags for the ACC, VEL, and DIS processed by Paolucci et al
    wf_tag_e_acc_mp = tag_e.replace('acc_cv','acc_mp')
    wf_tag_n_acc_mp = tag_n.replace('acc_cv','acc_mp')
    wf_tag_z_acc_mp = tag_z.replace('acc_cv','acc_mp')

    wf_tag_e_vel_mp = tag_e.replace('acc_cv','vel_mp')
    wf_tag_n_vel_mp = tag_n.replace('acc_cv','vel_mp')
    wf_tag_z_vel_mp = tag_z.replace('acc_cv','vel_mp')
        
    wf_tag_e_dis_mp = tag_e.replace('acc_cv','dis_mp')
    wf_tag_n_dis_mp = tag_n.replace('acc_cv','dis_mp')
    wf_tag_z_dis_mp = tag_z.replace('acc_cv','dis_mp')

    # update waveforms corrected by Paolucci et al
     #----------ACC----------*
    ds_mp_mb.add_waveforms(ds.waveforms[curr_sta.replace('.','_')][wf_tag_e_acc_mp], tag=wf_tag_e_acc_mp, event_id=evid)
    ds_mp_mb.add_waveforms(ds.waveforms[curr_sta.replace('.','_')][wf_tag_n_acc_mp], tag=wf_tag_n_acc_mp, event_id=evid)
    ds_mp_mb.add_waveforms(ds.waveforms[curr_sta.replace('.','_')][wf_tag_z_acc_mp], tag=wf_tag_z_acc_mp, event_id=evid)
  
    ds_mp_mb.add_auxiliary_data(data=zeros(0), data_type="Headers", path=pre_tag_path + wf_tag_e_acc_mp, parameters=ds.auxiliary_data.Headers[curr_sta.replace('.','_')][wf_tag_e_acc_mp].parameters)
    ds_mp_mb.add_auxiliary_data(data=zeros(0), data_type="Headers", path=pre_tag_path + wf_tag_n_acc_mp, parameters=ds.auxiliary_data.Headers[curr_sta.replace('.','_')][wf_tag_n_acc_mp].parameters)
    ds_mp_mb.add_auxiliary_data(data=zeros(0), data_type="Headers", path=pre_tag_path + wf_tag_z_acc_mp, parameters=ds.auxiliary_data.Headers[curr_sta.replace('.','_')][wf_tag_z_acc_mp].parameters)

    #----------VEL----------*
    ds_mp_mb.add_waveforms(ds.waveforms[curr_sta.replace('.','_')][wf_tag_e_vel_mp], tag=wf_tag_e_vel_mp, event_id=evid)
    ds_mp_mb.add_waveforms(ds.waveforms[curr_sta.replace('.','_')][wf_tag_n_vel_mp], tag=wf_tag_n_vel_mp, event_id=evid)
    ds_mp_mb.add_waveforms(ds.waveforms[curr_sta.replace('.','_')][wf_tag_z_vel_mp], tag=wf_tag_z_vel_mp, event_id=evid)
  
    ds_mp_mb.add_auxiliary_data(data=zeros(0), data_type="Headers", path=pre_tag_path + wf_tag_e_vel_mp, parameters=ds.auxiliary_data.Headers[curr_sta.replace('.','_')][wf_tag_e_vel_mp].parameters)
    ds_mp_mb.add_auxiliary_data(data=zeros(0), data_type="Headers", path=pre_tag_path + wf_tag_n_vel_mp, parameters=ds.auxiliary_data.Headers[curr_sta.replace('.','_')][wf_tag_n_vel_mp].parameters)
    ds_mp_mb.add_auxiliary_data(data=zeros(0), data_type="Headers", path=pre_tag_path + wf_tag_z_vel_mp, parameters=ds.auxiliary_data.Headers[curr_sta.replace('.','_')][wf_tag_z_vel_mp].parameters)    
    
    #----------DIS----------* 
    ds_mp_mb.add_waveforms(ds.waveforms[curr_sta.replace('.','_')][wf_tag_e_dis_mp], tag=wf_tag_e_dis_mp, event_id=evid)
    ds_mp_mb.add_waveforms(ds.waveforms[curr_sta.replace('.','_')][wf_tag_n_dis_mp], tag=wf_tag_n_dis_mp, event_id=evid)
    ds_mp_mb.add_waveforms(ds.waveforms[curr_sta.replace('.','_')][wf_tag_z_dis_mp], tag=wf_tag_z_dis_mp, event_id=evid)    
    
    ds_mp_mb.add_auxiliary_data(data=zeros(0), data_type="Headers", path=pre_tag_path + wf_tag_e_dis_mp, parameters=ds.auxiliary_data.Headers[curr_sta.replace('.','_')][wf_tag_e_dis_mp].parameters)
    ds_mp_mb.add_auxiliary_data(data=zeros(0), data_type="Headers", path=pre_tag_path + wf_tag_n_dis_mp, parameters=ds.auxiliary_data.Headers[curr_sta.replace('.','_')][wf_tag_n_dis_mp].parameters)
    ds_mp_mb.add_auxiliary_data(data=zeros(0), data_type="Headers", path=pre_tag_path + wf_tag_z_dis_mp, parameters=ds.auxiliary_data.Headers[curr_sta.replace('.','_')][wf_tag_z_dis_mp].parameters)     

    #----------SA-----------*
    ds_mp_mb.add_auxiliary_data(data=ds.auxiliary_data.Spectra[curr_sta.replace('.','_')][wf_tag_e_acc_mp].data, data_type="Spectra", path=pre_tag_path + wf_tag_e_acc_mp, parameters={'damping': 0.05, 'pga_cm_s_2': max(abs(ds.waveforms[curr_sta.replace('.','_')][wf_tag_e_acc_mp][0].data))})    
    ds_mp_mb.add_auxiliary_data(data=ds.auxiliary_data.Spectra[curr_sta.replace('.','_')][wf_tag_n_acc_mp].data, data_type="Spectra", path=pre_tag_path + wf_tag_n_acc_mp, parameters={'damping': 0.05, 'pga_cm_s_2': max(abs(ds.waveforms[curr_sta.replace('.','_')][wf_tag_n_acc_mp][0].data))})
    ds_mp_mb.add_auxiliary_data(data=ds.auxiliary_data.Spectra[curr_sta.replace('.','_')][wf_tag_z_acc_mp].data, data_type="Spectra", path=pre_tag_path + wf_tag_z_acc_mp, parameters={'damping': 0.05, 'pga_cm_s_2': max(abs(ds.waveforms[curr_sta.replace('.','_')][wf_tag_z_acc_mp][0].data))})

    #----------SD-----------*
    ds_mp_mb.add_auxiliary_data(data=ds.auxiliary_data.Spectra[curr_sta.replace('.','_')][wf_tag_e_dis_mp].data, data_type="Spectra", path=pre_tag_path + wf_tag_e_dis_mp, parameters={'damping': 0.05, 'pga_cm_s_2': max(abs(ds.waveforms[curr_sta.replace('.','_')][wf_tag_e_dis_mp][0].data))})    
    ds_mp_mb.add_auxiliary_data(data=ds.auxiliary_data.Spectra[curr_sta.replace('.','_')][wf_tag_n_dis_mp].data, data_type="Spectra", path=pre_tag_path + wf_tag_n_dis_mp, parameters={'damping': 0.05, 'pga_cm_s_2': max(abs(ds.waveforms[curr_sta.replace('.','_')][wf_tag_n_dis_mp][0].data))})
    ds_mp_mb.add_auxiliary_data(data=ds.auxiliary_data.Spectra[curr_sta.replace('.','_')][wf_tag_z_dis_mp].data, data_type="Spectra", path=pre_tag_path + wf_tag_z_dis_mp, parameters={'damping': 0.05, 'pga_cm_s_2': max(abs(ds.waveforms[curr_sta.replace('.','_')][wf_tag_z_dis_mp][0].data))})    
    
                       
    # define tags for the ACC, VEL, and DIS processed by eBASCO
    wf_tag_e_acc_ebasco = tag_e.replace('acc_cv','acc_mb')
    wf_tag_n_acc_ebasco = tag_n.replace('acc_cv','acc_mb')
    wf_tag_z_acc_ebasco = tag_z.replace('acc_cv','acc_mb')

    wf_tag_e_vel_ebasco = tag_e.replace('acc_cv','vel_mb')
    wf_tag_n_vel_ebasco = tag_n.replace('acc_cv','vel_mb')
    wf_tag_z_vel_ebasco = tag_z.replace('acc_cv','vel_mb')
        
    wf_tag_e_dis_ebasco = tag_e.replace('acc_cv','dis_mb')
    wf_tag_n_dis_ebasco = tag_n.replace('acc_cv','dis_mb')
    wf_tag_z_dis_ebasco = tag_z.replace('acc_cv','dis_mb')
    
    # update header of waveforms corrected by eBASCO
    head_e['baseline_correction'] = \
    head_n['baseline_correction'] = \
    head_z['baseline_correction'] = 'BASELINE REMOVED'
    
    head_e['high_cut_frequency_hz']=opts.he
    head_n['high_cut_frequency_hz']=opts.hn   
    head_z['high_cut_frequency_hz']=opts.hz
    head_e['filter_order'] = \
    head_n['filter_order'] = \
    head_z['filter_order'] = int(opts.fo)
    head_e['filter_type'] = \
    head_n['filter_type'] = \
    head_z['filter_type'] = 'BUTTERWORTH'
            
    head_e['processing_commandline']=commandLine
    head_n['processing_commandline']=commandLine
    head_z['processing_commandline']=commandLine
    
    head_e['seconds_to_cut_from_beginning']= START_CUT_E
    head_n['seconds_to_cut_from_beginning']= START_CUT_N
    head_z['seconds_to_cut_from_beginning']= START_CUT_Z 

    head_e['seconds_to_cut_from_end']= END_CUT_E
    head_n['seconds_to_cut_from_end']= END_CUT_N
    head_z['seconds_to_cut_from_end']= END_CUT_Z
            
    head_e['taper_at_beginning']= \
    head_n['taper_at_beginning']= \
    head_z['taper_at_beginning']= opts.ta 
    
    head_e['T1_sec']=T1_E_good
    head_n['T1_sec']=T1_N_good
    head_z['T1_sec']=T1_Z_good

    head_e['T2_sec']=T2_E_good
    head_n['T2_sec']=T2_N_good
    head_z['T2_sec']=T2_Z_good

    head_e['T3_sec']=T3_E_good
    head_n['T3_sec']=T3_N_good
    head_z['T3_sec']=T3_Z_good
            
    head_e['time_stamp'] = \
    head_n['time_stamp'] = \
    head_z['time_stamp'] = fromUTCDateTime(datetime.datetime.now())
    
    #----------ACC----------*
    if EW_pd != -999: ds_mp_mb.add_waveforms(st_E_acc, tag=tag_e[:-6] + "acc_" + proc_type, event_id=evid)
    if NS_pd != -999: ds_mp_mb.add_waveforms(st_N_acc, tag=tag_n[:-6] + "acc_" + proc_type, event_id=evid)
    if UD_pd != -999: ds_mp_mb.add_waveforms(st_Z_acc, tag=tag_z[:-6] + "acc_" + proc_type, event_id=evid)

    if EW_pd != -999: ds_mp_mb.add_auxiliary_data(data=zeros(0), data_type="Headers", path=pre_tag_path + wf_tag_e_acc_ebasco, parameters=head_e)
    if NS_pd != -999: ds_mp_mb.add_auxiliary_data(data=zeros(0), data_type="Headers", path=pre_tag_path + wf_tag_n_acc_ebasco, parameters=head_n)
    if UD_pd != -999: ds_mp_mb.add_auxiliary_data(data=zeros(0), data_type="Headers", path=pre_tag_path + wf_tag_z_acc_ebasco, parameters=head_z)
    #----------VEL----------*
    if EW_pd != -999: ds_mp_mb.add_waveforms(st_E_vel,tag=wf_tag_e_vel_ebasco, event_id=evid)
    if NS_pd != -999: ds_mp_mb.add_waveforms(st_N_vel,tag=wf_tag_n_vel_ebasco, event_id=evid)
    if UD_pd != -999: ds_mp_mb.add_waveforms(st_Z_vel,tag=wf_tag_z_vel_ebasco, event_id=evid)   

    head_e['units'] = head_e['units'] = head_e['units'] = 'cm/s'
    ds_mp_mb.add_auxiliary_data(data=zeros(0), data_type="Headers", path=pre_tag_path + wf_tag_e_vel_ebasco, parameters=head_e)
    ds_mp_mb.add_auxiliary_data(data=zeros(0), data_type="Headers", path=pre_tag_path + wf_tag_n_vel_ebasco, parameters=head_n)
    ds_mp_mb.add_auxiliary_data(data=zeros(0), data_type="Headers", path=pre_tag_path + wf_tag_z_vel_ebasco, parameters=head_z)

    #----------DIS----------*   
    if EW_pd != -999: ds_mp_mb.add_waveforms(st_E_dis,tag=wf_tag_e_dis_ebasco, event_id=evid)
    if NS_pd != -999: ds_mp_mb.add_waveforms(st_N_dis,tag=wf_tag_n_dis_ebasco, event_id=evid)
    if UD_pd != -999: ds_mp_mb.add_waveforms(st_Z_dis,tag=wf_tag_z_dis_ebasco, event_id=evid)   
    
    head_e['units'] = head_e['units'] = head_e['units'] = 'cm'
    if EW_pd != -999: ds_mp_mb.add_auxiliary_data(data=zeros(0), data_type="Headers", path=pre_tag_path + wf_tag_e_dis_ebasco, parameters=head_e)
    if NS_pd != -999: ds_mp_mb.add_auxiliary_data(data=zeros(0), data_type="Headers", path=pre_tag_path + wf_tag_n_dis_ebasco, parameters=head_n)
    if UD_pd != -999: ds_mp_mb.add_auxiliary_data(data=zeros(0), data_type="Headers", path=pre_tag_path + wf_tag_z_dis_ebasco, parameters=head_z)    
    
    #----------SA-----------*
    if EW_pd != -999: spe_E_acc = concatenate(([st_E_sa[0].data],[st_E_sa[1].data]),axis=0)
    if NS_pd != -999: spe_N_acc = concatenate(([st_N_sa[0].data],[st_N_sa[1].data]),axis=0)
    if UD_pd != -999: spe_Z_acc = concatenate(([st_Z_sa[0].data],[st_Z_sa[1].data]),axis=0)
    
    if EW_pd != -999: ds_mp_mb.add_auxiliary_data(data=spe_E_acc, data_type="Spectra", path=pre_tag_path + wf_tag_e_acc_ebasco, parameters={'damping': 0.05, 'pga_cm_s_2': st_E_acc[0].data[argmax(abs(st_E_acc[0].data))]})
    if NS_pd != -999: ds_mp_mb.add_auxiliary_data(data=spe_N_acc, data_type="Spectra", path=pre_tag_path + wf_tag_n_acc_ebasco, parameters={'damping': 0.05, 'pga_cm_s_2': st_N_acc[0].data[argmax(abs(st_N_acc[0].data))]})
    if UD_pd != -999: ds_mp_mb.add_auxiliary_data(data=spe_Z_acc, data_type="Spectra", path=pre_tag_path + wf_tag_z_acc_ebasco, parameters={'damping': 0.05, 'pga_cm_s_2': st_Z_acc[0].data[argmax(abs(st_Z_acc[0].data))]})

    #----------SD-----------*
    if EW_pd != -999: spe_E_dis = concatenate(([st_E_sd[0].data],[st_E_sd[1].data]),axis=0)
    if NS_pd != -999: spe_N_dis = concatenate(([st_N_sd[0].data],[st_N_sd[1].data]),axis=0)
    if UD_pd != -999: spe_Z_dis = concatenate(([st_Z_sd[0].data],[st_Z_sd[1].data]),axis=0)

    if EW_pd != -999: ds_mp_mb.add_auxiliary_data(data=spe_E_dis, data_type="Spectra", path=pre_tag_path + wf_tag_e_dis_ebasco, parameters={'damping': 0.05, 'pgd_cm': st_E_dis[0].data[argmax(abs(st_E_dis[0].data))]})
    if NS_pd != -999: ds_mp_mb.add_auxiliary_data(data=spe_N_dis, data_type="Spectra", path=pre_tag_path + wf_tag_n_dis_ebasco, parameters={'damping': 0.05, 'pgd_cm': st_N_dis[0].data[argmax(abs(st_N_dis[0].data))]})
    if UD_pd != -999: ds_mp_mb.add_auxiliary_data(data=spe_Z_dis, data_type="Spectra", path=pre_tag_path + wf_tag_z_dis_ebasco, parameters={'damping': 0.05, 'pgd_cm': st_Z_dis[0].data[argmax(abs(st_Z_dis[0].data))]})



    #add quakeml and stationml if present
    quakeml_filename = opts.ifile[:-3]+"_quake.xml"
    if os_path_isfile(quakeml_filename): ds_new.add_quakeml(quakeml_filename)
    stationml_filename = opts.ifile[:-3] + "_station.xml"
    if os_path_isfile(stationml_filename): ds_new.add_stationxml(stationml_filename)

    if opts.verbose:
        print ""
        print "CMD line: " + commandLine
        print ""
        print "--------o--------o--------o--------o--------o--------o--------o--------"
        print ""
                                                        
    end = time.time()
    print("processing time: " , format(end - start, '.3f'),"s")
    print("")
    print("#########################################")
    print("")    
    print(commandLine)
    print("") 
    print("#########################################")
    
    return dict_fmt(exit_status=0, in_file=opts.ifile, out_file=ds_mp_mb.filename, out_format='hdf5') 
    # ricordarsi di modificare out_file se si cambia asdf file (BASCO -->ds_new/ ESM+BASCO -->ds_mp_mb
    
#-------------------------------------------------------------#          
# 							FUNCTIONS                         #
#-------------------------------------------------------------#

# read waveforms tags in the ASDF file
def read_wf_tags(wf_tags_list):
    for curr_tag in wf_tags_list:
        if (('e_' in curr_tag[0:len(curr_tag)])==1 | ('1_' in curr_tag[0:len(curr_tag)])==1) & ('_acc_cv' in curr_tag[0:len(curr_tag)])==1:
            tag_E_RAW = curr_tag
            tag_E_CORR = tag_E_RAW.replace("_acc_cv","_acc_mp")
        elif ('n_' in curr_tag[0:len(curr_tag)])==1 | ('2_' in curr_tag[0:len(curr_tag)])==1 & ('_acc_cv' in curr_tag[0:len(curr_tag)])==1:
            tag_N_RAW = curr_tag
            tag_N_CORR = tag_N_RAW.replace("_acc_cv","_acc_mp")
        elif ('z_' in curr_tag[0:len(curr_tag)])==1 & ('_acc_cv' in curr_tag[0:len(curr_tag)])==1:
            tag_Z_RAW = curr_tag
            tag_Z_CORR = tag_Z_RAW.replace("_acc_cv","_acc_mp")
    
            wf_tags_dct = {"tag_E_RAW":tag_E_RAW,"tag_E_CORR":tag_E_CORR,"tag_N_RAW":tag_N_RAW,"tag_N_CORR":tag_N_CORR,"tag_Z_RAW":tag_Z_RAW,"tag_Z_CORR":tag_Z_CORR}
            return wf_tags_dct

# slice
def wf_Slice(st_a, st_b, st_c, c_a, c_z):
    max_starttime = max(st_a[0].stats.starttime, st_b[0].stats.starttime, st_c[0].stats.starttime)
    new_max_starttime = max_starttime + c_a
    #if max_starttime.microsecond > 0: max_starttime = max_starttime + (1.0 - max_starttime.microsecond/1000000.0)
    min_endtime = min(st_a[0].stats.endtime, st_b[0].stats.endtime, st_c[0].stats.endtime)
    if c_z == 0:
        new_min_endtime = min_endtime
    else:
        new_min_endtime = max_starttime + c_z
#    if min_endtime.microsecond > 0: min_endtime = min_endtime - min_endtime.microsecond/1000000.0
    st_a[0].trim(starttime=new_max_starttime, endtime=new_min_endtime, pad=True, fill_value=st_a[0][0])
    st_b[0].trim(starttime=new_max_starttime, endtime=new_min_endtime, pad=True, fill_value=st_b[0][0])
    st_c[0].trim(starttime=new_max_starttime, endtime=new_min_endtime, pad=True, fill_value=st_c[0][0])
    # print(max_starttime-new_max_starttime)
    # print(min_endtime-new_min_endtime)
    try: 
        st_a[0].data = st_a[0].data.filled()
        st_b[0].data = st_b[0].data.filled()
        st_c[0].data = st_c[0].data.filled()
    except AttributeError: pass

    # next line fixes faulty approx made the by obspy slice function
    #st_a[0].stats.starttime = st_b[0].stats.starttime = st_c[0].stats.starttime = max_starttime
    return st_a, st_b, st_c

# converts UTCDateTime object into a string
def fromUTCDateTime(dt):
    # dt = :class:`~obspy.core.UTCDateTime`
    pattern = "%4d%02d%02d_%02d%02d%02d.%03d"
    return pattern % (dt.year, dt.month, dt.day, dt.hour,
                        dt.minute, dt.second, dt.microsecond / 1000)
                                
# calculate the duration of the strong ground motion by ARIAS intensity criterion
def wf_Partition(w_f, d_t, i1, i2):
    N = len(w_f)
    Ia = zeros(N)
    for i in range(0,N-1):
        Ia[i+1] = Ia[i] + w_f[i+1]**2
    Ia = Ia/Ia[-1]
    idx1 = (abs(Ia-i1)).argmin()
    idx2 = (abs(Ia-i2)).argmin()
    return idx1*d_t,(idx2)*d_t,(idx2-idx1)*d_t

# read raw and corrected waveforms and related metadata in the ASDF file     
def wf_md_Extract(ds,sta,wf_raw_tag,wf_corr_tag):    
    ACC_RAW = ds.waveforms[sta][wf_raw_tag][0].data
    ACC_CORR = ds.waveforms[sta][wf_corr_tag][0].data
    #eve_id = ds.auxiliary_data.Headers[sta.replace('.','_')][wf_raw_tag].parameters['event_id']
    # eve_id = str(ds.auxiliary_data.Headers[sta][wf_raw_tag].parameters['event_id'])
    starttime = ds.waveforms[sta][wf_raw_tag][0].stats["starttime"]
    endtime = ds.waveforms[sta][wf_raw_tag][0].stats["endtime"]
    sampling_rate = ds.waveforms[sta][wf_raw_tag][0].stats["delta"]
    wf_dct = {"ACC_RAW":ACC_RAW,"ACC_CORR":ACC_CORR,"starttime":starttime,"endtime":endtime,"sampling_rate":sampling_rate}
    return wf_dct
    
# duplicates streams
def dpcpy_Stream(st_a, st_b, st_c):
    st_A = copy_deepcopy(st_a); 
    st_B = copy_deepcopy(st_b); 
    st_C = copy_deepcopy(st_c); 
    return st_A, st_B, st_C

# integrate the acceleration to obtain velocity and displacement
def double_integ(AMP,dt):
    TIME = arange(0,(len(AMP)-0.5)*dt,dt)
    # create temporary arrays for integration
    strms_AMP = copy_deepcopy(AMP); # make a copy of AMP
    ACC_tmp = strms_AMP;
    VEL = cumsum(ACC_tmp)*dt
    VEL_copy = copy_deepcopy(VEL);
    DIS = cumsum(VEL_copy)*dt
    return VEL,DIS,TIME
    
# integrate the acceleration to obtain velocity and displacement
def double_integ_strm(AMP,dt):
    TIME = arange(0,(len(AMP)-0.5)*dt,dt)
    # create temporary arrays for integration
    strms_AMP = copy_deepcopy(AMP) # make a copy of AMP
    VEL_tmp = strms_AMP.integrate()
    VEL = copy_deepcopy(VEL_tmp)
    DIS = VEL_tmp.integrate()
    return VEL,DIS,TIME

# cut waveforms by energy criterion (Arias Intensity) 
def cut_wf_ARIAS (wf_E,wf_N,wf_Z,dt_E,dt_N,dt_Z,P_start,P_end,optsca,optscz,mfst,mfnd):
    ACC_E_strm = copy_deepcopy(wf_E)
    ACC_N_strm = copy_deepcopy(wf_N)
    ACC_Z_strm = copy_deepcopy(wf_Z)   
    
    if optsca==0 and optscz==0:
        # cut acceleration according to energy criterion       
        ARIAS_E = wf_Partition(wf_E[0].data,dt_E, P_start, P_end); ACC_E_strm.trim(starttime=wf_E[0].stats.starttime+ARIAS_E[0]-mfst*ARIAS_E[2], endtime=wf_E[0].stats.starttime+ARIAS_E[1]+mfnd*ARIAS_E[2], pad=True, fill_value=ACC_E_strm[0][0])  
        
        ARIAS_N = wf_Partition(wf_N[0].data,dt_N, P_start, P_end); ACC_N_strm.trim(starttime=wf_N[0].stats.starttime+ARIAS_N[0]-mfst*ARIAS_N[2], endtime=wf_N[0].stats.starttime+ARIAS_N[1]+mfnd*ARIAS_N[2], pad=True, fill_value=ACC_N_strm[0][0])
        
        ARIAS_Z = wf_Partition(wf_Z[0].data,dt_Z, P_start, P_end); ACC_Z_strm.trim(starttime=wf_Z[0].stats.starttime+ARIAS_Z[0]-mfst*ARIAS_Z[2], endtime=wf_Z[0].stats.starttime+ARIAS_Z[1]+mfnd*ARIAS_Z[2], pad=True, fill_value=ACC_Z_strm[0][0])
        
        cut_start_E = ARIAS_E[0]-mfst*ARIAS_E[2]
        cut_start_N = ARIAS_N[0]-mfst*ARIAS_N[2]
        cut_start_Z = ARIAS_Z[0]-mfst*ARIAS_Z[2]

        cut_end_E   = ARIAS_E[1]+mfnd*ARIAS_E[2]
        cut_end_N   = ARIAS_N[1]+mfnd*ARIAS_N[2]
        cut_end_Z   = ARIAS_Z[1]+mfnd*ARIAS_Z[2]
        cut_start=[cut_start_E,cut_start_N,cut_start_Z]
        cut_end=[cut_end_E,cut_end_N,cut_end_Z]

    else:
        cut_start = [optsca,optsca,optsca]
        cut_end = [optscz, optscz,optscz]
    return [ACC_E_strm, ACC_N_strm, ACC_Z_strm,cut_start, cut_end]

def ACCtoDISP (ACC_E_strm, ACC_N_strm, ACC_Z_strm,dt_E,dt_N,dt_Z):
	
    ACC_E = ACC_E_strm[0].data - ACC_E_strm[0][0]
    ACC_E_copy = copy_deepcopy(ACC_E)    
    [VEL_E, DIS_E, TIME_E] = double_integ(ACC_E_copy,dt_E) 
       
    ACC_N = ACC_N_strm[0].data - ACC_N_strm[0][0]
    ACC_N_copy = copy_deepcopy(ACC_N)    
    [VEL_N, DIS_N, TIME_N] = double_integ(ACC_N_copy,dt_N)  

    ACC_Z = ACC_Z_strm[0].data - ACC_Z_strm[0][0]
    ACC_Z_copy = copy_deepcopy(ACC_Z)    
    [VEL_Z, DIS_Z, TIME_Z] = double_integ(ACC_Z_copy,dt_Z)  
    
    str_TIME = [TIME_E, TIME_N, TIME_Z ]
    str_ACC = [ACC_E_copy, ACC_N_copy, ACC_Z_copy]
    str_VEL = [VEL_E, VEL_N, VEL_Z]
    str_DIS = [DIS_E, DIS_N, DIS_Z]
    
    return [str_ACC,str_VEL,str_DIS, str_TIME]
    
# cut waveforms by manually set starttime and endtime 
def cut_wf (wf_E,wf_N,wf_Z,start_E,start_N,start_Z,end_E,end_N,end_Z):              
    E_str = copy_deepcopy(wf_E)    
    E_str[0].trim(starttime=E_str[0].stats.starttime+start_E, endtime=E_str[0].stats.starttime+end_E)
    N_str = copy_deepcopy(wf_N)
    N_str[0].trim(starttime=E_str[0].stats.starttime+start_N, endtime=E_str[0].stats.starttime+end_N)
    Z_str = copy_deepcopy(wf_Z)
    Z_str[0].trim(starttime=E_str[0].stats.starttime+start_Z, endtime=E_str[0].stats.starttime+end_Z)
    return(E_str,N_str,Z_str)
                
    
# sample the T1 time before the 5% of the cumulated energy (ARIAS INTENSITY)
def sample_T1 (wf_E,wf_N,wf_Z,dt_E,dt_N,dt_Z,opts_t1):
    T1_range_E = wf_Partition(wf_E, dt_E, 0.0001, 0.05)
    T1_range_N = wf_Partition(wf_N, dt_N, 0.0001, 0.05)
    T1_range_Z = wf_Partition(wf_Z, dt_Z, 0.0001, 0.05)
    T1_sample_E = logspace(log10(T1_range_E[0]),log10(T1_range_E[1]),opts_t1, endpoint=False, base=10.0)
    T1_sample_N = logspace(log10(T1_range_N[0]),log10(T1_range_N[1]),opts_t1, endpoint=False, base=10.0)
    T1_sample_Z = logspace(log10(T1_range_Z[0]),log10(T1_range_Z[1]),opts_t1, endpoint=False, base=10.0)
    return T1_sample_E, T1_sample_N, T1_sample_Z
    
# sample T3 time after the 5% of the cumulated energy (ARIAS INTENSITY)
def sample_T3 (wf_E,wf_N,wf_Z,dt_E,dt_N,dt_Z,opts_t3):
    T3_range_E = wf_Partition(wf_E, dt_E, 0.05, 0.95)
    T3_range_N = wf_Partition(wf_N, dt_N, 0.05, 0.95)
    T3_range_Z = wf_Partition(wf_Z, dt_Z, 0.05, 0.95)
    T3_sample_E = logspace(log10(T3_range_E[0]),log10(T3_range_E[1]),opts_t3, endpoint=False, base=10.0)
    T3_sample_N = logspace(log10(T3_range_N[0]),log10(T3_range_N[1]),opts_t3, endpoint=False, base=10.0)
    T3_sample_Z = logspace(log10(T3_range_Z[0]),log10(T3_range_Z[1]),opts_t3, endpoint=False, base=10.0)
    return T3_sample_E, T3_sample_N, T3_sample_Z
    
# search the end point for each component
def end_point (wf_E,wf_N,wf_Z,dt_E,dt_N,dt_Z):
    #T3_range_E = wf_Partition(wf_E, dt_E, 0.05, 0.95)
    #T3_range_N = wf_Partition(wf_N, dt_N, 0.05, 0.95)
    #T3_range_Z = wf_Partition(wf_Z, dt_Z, 0.05, 0.95)
    #return  T3_range_E[1],T3_range_N[1],T3_range_Z[1] 
    time_E = arange(0,len(wf_E)*dt_E,dt_E) 
    time_N = arange(0,len(wf_N)*dt_N,dt_N)
    time_Z = arange(0,len(wf_Z)*dt_E,dt_Z)
    END_POINTS = [time_E[-1], time_N[-1], time_Z [-1]]
    return END_POINTS
    
# sample T2 time from each T3 and the end of the signal
def T2_sample(T3_SAMPLES,END_POINTS,ind,opts_t2):
    T2_SAMP_E = []
    T2_SAMP_N = []
    T2_SAMP_Z = []
    
    T2_SAMP_E=logspace(log10(T3_SAMPLES[0][ind]),log10(END_POINTS[0]), opts_t2, endpoint=False, base=10.0)
    T2_SAMP_N=logspace(log10(T3_SAMPLES[1][ind]),log10(END_POINTS[1]), opts_t2, endpoint=False, base=10.0)       
    T2_SAMP_Z=logspace(log10(T3_SAMPLES[2][ind]),log10(END_POINTS[2]), opts_t2, endpoint=False, base=10.0)

    return T2_SAMP_E, T2_SAMP_N, T2_SAMP_Z
        
# find index of the closest values in numpy array
def find_nearest(array, value):
    array = asarray(array)
    idx = (abs(array - value)).argmin()
    return idx

# define the pre-event
def pre_eve(wf_E,wf_N,wf_Z,dt_E,dt_N,dt_Z,T1_ARRAY):
    time_E = arange(0,len(wf_E)*dt_E,dt_E)
    PRE_EV_TIME_E =[]
    PRE_EV_VEL_E = []
    for T1 in T1_ARRAY[0]:
        PRE_EV_TIME_E.append(time_E[0:find_nearest(time_E, T1)])
        PRE_EV_VEL_E.append(wf_E[0:find_nearest(time_E, T1)])
    time_N = arange(0,len(wf_N)*dt_N,dt_N)
    PRE_EV_TIME_N =[]
    PRE_EV_VEL_N = []
    for T1 in T1_ARRAY[1]:
        PRE_EV_TIME_N.append(time_N[0:find_nearest(time_N, T1)])
        PRE_EV_VEL_N.append(wf_N[0:find_nearest(time_N, T1)])
    time_Z = arange(0,len(wf_Z)*dt_Z,dt_Z)
    PRE_EV_TIME_Z =[]
    PRE_EV_VEL_Z = []
    for T1 in T1_ARRAY[2]:
        PRE_EV_TIME_Z.append(time_Z[0:find_nearest(time_Z, T1)])
        PRE_EV_VEL_Z.append(wf_Z[0:find_nearest(time_Z, T1)])
    PRE_EVE_TIME = [PRE_EV_TIME_E,PRE_EV_TIME_N,PRE_EV_TIME_Z]
    PRE_EVE_VEL = [PRE_EV_VEL_E,PRE_EV_VEL_N,PRE_EV_VEL_Z]
    return PRE_EVE_TIME, PRE_EVE_VEL 

def func_pre_fit(x, a):       
    return a*x 
        
# de-trend the pre-event func = ax
def pre_eve_detrend (PRE_TIME,PRE_VEL):
    PRE_EVE_VEL_DET_E = []
    PRE_EVE_VEL_DET_N = []
    PRE_EVE_VEL_DET_Z = []
    LAST_VAL_FIT_E = []
    LAST_VAL_FIT_N = []
    LAST_VAL_FIT_Z = []
    FIT_E = []
    FIT_N = []
    FIT_Z = []
        
    for ind in  arange(len(PRE_VEL[0])):
        popt1,pcov1 = optimize.curve_fit(func_pre_fit,PRE_TIME[0][ind],PRE_VEL[0][ind])  
        PRE_EVE_VEL_DET_E.append(PRE_VEL[0][ind]- func_pre_fit(PRE_TIME[0][ind],popt1))
        LAST_VAL_FIT_E.append(func_pre_fit(PRE_TIME[0][ind],popt1)[-1])
        FIT_E.append(func_pre_fit(PRE_TIME[0][ind],popt1))
    for ind in  arange(len(PRE_VEL[1])):
        popt2,pcov2 = optimize.curve_fit(func_pre_fit,PRE_TIME[1][ind],PRE_VEL[1][ind])  
        PRE_EVE_VEL_DET_N.append(PRE_VEL[1][ind]- func_pre_fit(PRE_TIME[1][ind],popt2))
        LAST_VAL_FIT_N.append(func_pre_fit(PRE_TIME[1][ind],popt2)[-1])
        FIT_N.append(func_pre_fit(PRE_TIME[1][ind],popt1))
    for ind in  arange(len(PRE_VEL[2])):
        popt3,pcov3 = optimize.curve_fit(func_pre_fit,PRE_TIME[2][ind],PRE_VEL[2][ind])  
        PRE_EVE_VEL_DET_Z.append(PRE_VEL[2][ind]- func_pre_fit(PRE_TIME[2][ind],popt3))
        LAST_VAL_FIT_Z.append(func_pre_fit(PRE_TIME[2][ind],popt3)[-1])
        FIT_Z.append(func_pre_fit(PRE_TIME[2][ind],popt1))
        
    PRE_EVE_VEL_DET = [PRE_EVE_VEL_DET_E, PRE_EVE_VEL_DET_N, PRE_EVE_VEL_DET_Z]
    LAST_VALUE_LIN_FIT = [LAST_VAL_FIT_E,LAST_VAL_FIT_N,LAST_VAL_FIT_Z]
    LIN_FIT = [FIT_E,FIT_N,FIT_Z]
    
    return PRE_EVE_VEL_DET, LAST_VALUE_LIN_FIT, LIN_FIT

# linear de-trend 
def lin_det(tr1,tr2,tr3):
    PRE_EVE_VEL_DETRENDED_E = []
    for val in arange(len(tr1)):
        PRE_EVE_VEL_DETRENDED_E.append(detrend_linear(tr1[val]))
    PRE_EVE_VEL_DETRENDED_N = []
    for val in arange(len(tr2)):
        PRE_EVE_VEL_DETRENDED_N.append(detrend_linear(tr2[val]))
    PRE_EVE_VEL_DETRENDED_Z = []
    for val in arange(len(tr3)):
        PRE_EVE_VEL_DETRENDED_Z.append(detrend_linear(tr3[val]))
    PRE_EVE_DETR = [PRE_EVE_VEL_DETRENDED_E,PRE_EVE_VEL_DETRENDED_N,PRE_EVE_VEL_DETRENDED_Z]
    return PRE_EVE_DETR
    
# define the post-event
def post_eve(wf_E,wf_N,wf_Z,dt_E,dt_N,dt_Z,T2_ARRAY,END_POINTS,index_T2):
    time_E = arange(0,len(wf_E)*dt_E,dt_E)
    time_N = arange(0,len(wf_N)*dt_N,dt_N)
    time_Z = arange(0,len(wf_Z)*dt_Z,dt_Z)

    POST_EV_TIME_E=time_E[find_nearest(time_E, T2_ARRAY[0][index_T2]):find_nearest(time_E, END_POINTS[0])]
    POST_EV_VEL_E=wf_E[find_nearest(time_E, T2_ARRAY[0][index_T2]):find_nearest(time_E, END_POINTS[0])]
    
    POST_EV_TIME_N=time_N[find_nearest(time_N, T2_ARRAY[1][index_T2]):find_nearest(time_N, END_POINTS[1])]
    POST_EV_VEL_N=wf_N[find_nearest(time_N, T2_ARRAY[1][index_T2]):find_nearest(time_N, END_POINTS[1])]
    
    POST_EV_TIME_Z=time_Z[find_nearest(time_Z, T2_ARRAY[2][index_T2]):find_nearest(time_Z, END_POINTS[2])]
    POST_EV_VEL_Z=wf_Z[find_nearest(time_Z, T2_ARRAY[2][index_T2]):find_nearest(time_Z, END_POINTS[2])]
           
    POST_EVE_TIME = [POST_EV_TIME_E,POST_EV_TIME_N,POST_EV_TIME_Z]
    POST_EVE_VEL = [POST_EV_VEL_E,POST_EV_VEL_N,POST_EV_VEL_Z]
    return POST_EVE_TIME, POST_EVE_VEL
      
def func_post_fit(x, b, c):       
    return b*x +c        
        
# de-trend the post-event func = bx+c
def post_eve_detrend (POST_TIME,POST_VEL):
        
    # ~ for ind in  arange(len(POST_VEL[0])):
    popt1,pcov1 = optimize.curve_fit(func_post_fit,POST_TIME[0],POST_VEL[0])  
    POST_EVE_VEL_DET_E=POST_VEL[0]- func_post_fit(POST_TIME[0],popt1[0],popt1[1])
    # FIRST_VAL_FIT_E=func_post_fit(POST_TIME[0],popt1[0],popt1[1])[-1]
    FIT_E=func_post_fit(POST_TIME[0],popt1[0],popt1[1])
    # ~ for ind in  arange(len(POST_VEL[1])):
    popt2,pcov2 = optimize.curve_fit(func_post_fit,POST_TIME[1],POST_VEL[1])  
    POST_EVE_VEL_DET_N=POST_VEL[1]- func_post_fit(POST_TIME[1],popt2[0],popt2[1])
    # FIRST_VAL_FIT_N=func_post_fit(POST_TIME[1],popt2[0],popt2[1])[-1]
    FIT_N=func_post_fit(POST_TIME[1],popt2[0],popt2[1])
    # ~ for ind in  arange(len(POST_VEL[2])):
    popt3,pcov3 = optimize.curve_fit(func_post_fit,POST_TIME[2],POST_VEL[2])  
    POST_EVE_VEL_DET_Z=POST_VEL[2]- func_post_fit(POST_TIME[2],popt3[0],popt3[1])
    # FIRST_VAL_FIT_Z=func_post_fit(POST_TIME[2],popt3[0],popt3[1])[-1]
    FIT_Z=func_post_fit(POST_TIME[2],popt3[0],popt3[1])
        
    POST_EVE_VEL_DET = [POST_EVE_VEL_DET_E, POST_EVE_VEL_DET_N, POST_EVE_VEL_DET_Z]
    # FIRST_VALUE_LIN_FIT = [FIRST_VAL_FIT_E,FIRST_VAL_FIT_N,FIRST_VAL_FIT_Z]
    LIN_FIT = [FIT_E,FIT_N,FIT_Z]
    
    return POST_EVE_VEL_DET, LIN_FIT

# define the strong-event
def strong_eve(wf_E,wf_N,wf_Z,dt_E,dt_N,dt_Z,T1,T2,ind_1,ind_2):
    time_E = arange(0,len(wf_E)*dt_E,dt_E)
    time_N = arange(0,len(wf_E)*dt_N,dt_N)
    time_Z = arange(0,len(wf_Z)*dt_Z,dt_Z)
    
    STRONG_EV_TIME_E = time_E[find_nearest(time_E, T1[0][ind_1]):find_nearest(time_E, T2[0][ind_2])]
    STRONG_EV_VEL_E = wf_E[find_nearest(time_E, T1[0][ind_1]):find_nearest(time_E, T2[0][ind_2])]
    STRONG_EV_TIME_N = time_N[find_nearest(time_N, T1[1][ind_1]):find_nearest(time_N, T2[1][ind_2])]
    STRONG_EV_VEL_N = wf_N[find_nearest(time_N, T1[1][ind_1]):find_nearest(time_N, T2[1][ind_2])]
    STRONG_EV_TIME_Z = time_Z[find_nearest(time_Z, T1[2][ind_1]):find_nearest(time_Z, T2[2][ind_2])]
    STRONG_EV_VEL_Z = wf_Z[find_nearest(time_Z, T1[2][ind_1]):find_nearest(time_Z, T2[2][ind_2])]
    
    STRONG_EVE_TIME = [STRONG_EV_TIME_E,STRONG_EV_TIME_N,STRONG_EV_TIME_Z]
    STRONG_EVE_VEL = [STRONG_EV_VEL_E,STRONG_EV_VEL_N,STRONG_EV_VEL_Z]
    return STRONG_EVE_TIME, STRONG_EVE_VEL

def line_eq(x,m,q):
    return m*x + q
    
# define the line between T1 and T2 to be subtracted from the velocity
def strong_eve_line (STRONG_PHASE_TIME,Am_E,Am_N,Am_Z,q_E,q_N,q_Z):
    STRONG_PHASE_LINE_E = retta(STRONG_PHASE_TIME[0],Am_E,q_E)
    STRONG_PHASE_LINE_N = retta(STRONG_PHASE_TIME[1],Am_N,q_N)
    STRONG_PHASE_LINE_Z = retta(STRONG_PHASE_TIME[2],Am_Z,q_Z)
                
    return STRONG_PHASE_LINE_E, STRONG_PHASE_LINE_N, STRONG_PHASE_LINE_Z
    
# de-trend the strong-event
def strong_eve_detrend (STRONG_TIME,STRONG_VEL,STRONG_EVE_LINE):
    STRONG_EVE_VEL_DET_E = STRONG_VEL[0]- STRONG_EVE_LINE[0]
    STRONG_EVE_VEL_DET_N = STRONG_VEL[1]- STRONG_EVE_LINE[1]
    STRONG_EVE_VEL_DET_Z = STRONG_VEL[2]- STRONG_EVE_LINE[2]
        
    STRONG_EVE_VEL_DET = [STRONG_EVE_VEL_DET_E, STRONG_EVE_VEL_DET_N, STRONG_EVE_VEL_DET_Z]
    
    return STRONG_EVE_VEL_DET                        
    
# find corrected acceleration compatible with the uncorrected acceleration
def acc_corr (st_acc,st_acc_corr,dt,t1,t2,t3,opts_eps):
    ACC_CORR_OK = []
    ACC_CORR_KO = []
    T3_OK = []
    INDEX_OK = []

    for index in arange(len(st_acc_corr)):
        index_1 = find_nearest(arange(0,len(st_acc_corr[index])*dt,dt), t1[index]) 
        PGA_UNCORR_T1 = st_acc[index_1]
        PGA_CORR_T1 = st_acc_corr[index][index_1]
    
        index_2 = find_nearest(arange(0,len(st_acc_corr[index])*dt,dt), t2[index]) 
        PGA_UNCORR_T2 = st_acc[index_2]
        PGA_CORR_T2 = st_acc_corr[index][index_2]
    
        p_var_t1 = abs(PGA_CORR_T1-PGA_UNCORR_T1)/abs(PGA_UNCORR_T1) # variazione percentuale rispetto al valore del non corretto
        p_var_t2 = abs(PGA_CORR_T2-PGA_UNCORR_T2)/abs(PGA_UNCORR_T2)  
    
        # print('p_var_t1:' + str(p_var_t1) + '-----' + 'p_var_t2:' + str(p_var_t2))
        
        if p_var_t1 < opts_eps and p_var_t2 < opts_eps:
        # if p_var_t2 < epsilon:   
            ACC_CORR_OK.append(st_acc_corr[index])
            # ~ print('p_var_t1:' + str(p_var_t1) + '-----' + 'p_var_t2:' + str(p_var_t2))
            T3_OK.append(t3[index])
            INDEX_OK.append(index)
        else:  
            ACC_CORR_KO.append(st_acc_corr[index]) 

    return ACC_CORR_OK,ACC_CORR_KO,T3_OK,INDEX_OK
                
# acceptable solutions
def accept_solution(acc_corr, dt, comp_str):
    time_corr_good = []
    vel_corr_good = []
    disp_corr_good = []
    if not(acc_corr): print("nessuna soluzione per " + comp_str)          		
    else: 
        # double integrate the good acceleration to obtaine good velocity and displacement  

        strms_cp = copy_deepcopy(acc_corr)
        for ind in arange(len(acc_corr)):
            OUT = double_integ(strms_cp[ind],dt)
            time_corr_good.append(OUT[2])
            vel_corr_good.append(OUT[0])
            disp_corr_good.append(OUT[1])	       
    return (time_corr_good,vel_corr_good, disp_corr_good)                    		
   
# calculate the f-value 
def fvalue (time,amp,t3):
    f_value = [];
    for num in arange(len(amp)):
        # time = arange(0,len(amp[num])*dt,dt)
        ind_T3 = find_nearest(time[num],t3[num])
        slope, intercept, r_value, p_value, std_err = stats.linregress(time[num][ind_T3:-1],amp[num][ind_T3:-1])
        sigma = std_err**2
        f_value.append(abs(r_value)/(abs(slope)*sigma))
    return f_value 

# data taper at the beginning
def AsymTaper_beginning(w_f, n_pts, t_a):
    tap1 = cosTaper(n_pts, t_a*2*10**-2)
    tap = ones(n_pts)
    tap[:int(n_pts/2)] = tap1[:int(n_pts/2)]
    return w_f*tap  

def final_trace(ACC_CORR_X_good,num_X,t_a,std_delta_X,f_o,l_f):
    # taper at the beginning of the signal (default 5%)
    strms_X = AsymTaper_beginning(ACC_CORR_X_good[num_X], len(ACC_CORR_X_good[num_X]), t_a)
        
    # Low-pass acausal Butterworth filter
    strms_X_fltrd = lowpass(strms_X,l_f, int(1/std_delta_X), corners=f_o, zerophase=True)
        
    # create temporary arrays for differentiation
    data_X_temporary = copy_deepcopy(strms_X_fltrd)
        
    # VELOCITY
    #---------------------------------------------------------------

    data_X_temporary = cumsum(data_X_temporary)*std_delta_X

    # taper velocity
    data_X_temporary = AsymTaper_beginning(data_X_temporary, len(data_X_temporary) , t_a)
		
    # DISPLACEMENTS
    #---------------------------------------------------------------

    data_X_temporary = cumsum(data_X_temporary)*std_delta_X

    # taper displacement
    st_X_dis = AsymTaper_beginning(data_X_temporary, len(data_X_temporary) , t_a)

    # reconstruct Trace object
    #---------------------------

    # back to velocity
    (st_X_vel) = copy_deepcopy(st_X_dis) # make a copy of streams 
    st_X_vel = gradient(st_X_vel,std_delta_X)

    # back to acceleration
    (st_X_acc) = copy_deepcopy(st_X_vel) # make a copy of streams 
    st_X_acc = gradient(st_X_acc,std_delta_X)
    
    return st_X_acc,st_X_vel,st_X_dis
                    
def dict_fmt(exit_status, exit_message=None, in_file=None, out_string=None, out_file=None, out_format=None, out_file_content=None):
    return { \
    'exit_status': exit_status, \
    'exit_message': exit_message, \
    'in_file': in_file, \
    'out_string': out_string, \
    'out_file': out_file, \
    'out_format': out_format, \
    'out_file_content': out_file_content
    }                                             

def update_ebasco_trace (curr_sta,tag_comp,ds,starttime,endtime,ACC_X_eb,VEL_X_eb,DIS_X_eb):

    # define the tags for the ACC, VEL, and DIS time histories processed by eBASCO
    wf_tag_x_acc_ebasco = tag_comp.replace('acc_cv','acc_mb')

    
    wf_tag_x_vel_ebasco = tag_comp.replace('acc_cv','vel_mb')

    
    wf_tag_e_dis_ebasco = tag_comp.replace('acc_cv','dis_mb')
  
    
    # update the ASDF file with the ACC, VEL, and DIS time histories processed by eBASCO
    
    # update stats
    trc_x_acc_ebasco = trace.Trace(data=array(ACC_X_eb))
    trc_x_vel_ebasco = trace.Trace(data=array(VEL_X_eb))
    trc_x_dis_ebasco = trace.Trace(data=array(DIS_X_eb))
    
    trc_x_acc_ebasco.stats.network = ds.waveforms[curr_sta.replace('.','_')][tag_comp][0].stats.network
    trc_x_acc_ebasco.stats.station = ds.waveforms[curr_sta.replace('.','_')][tag_comp][0].stats.station
    trc_x_acc_ebasco.stats.location = ds.waveforms[curr_sta.replace('.','_')][tag_comp][0].stats.location
    trc_x_acc_ebasco.stats.channel = ds.waveforms[curr_sta.replace('.','_')][tag_comp][0].stats.channel
    trc_x_acc_ebasco.stats.starttime = starttime
    trc_x_acc_ebasco.stats.sampling_rate = ds.waveforms[curr_sta.replace('.','_')][tag_comp][0].stats.sampling_rate
    trc_x_acc_ebasco.stats.delta = ds.waveforms[curr_sta.replace('.','_')][tag_comp][0].stats.delta
    trc_x_acc_ebasco.stats.npts = len(ACC_X_eb)
    trc_x_acc_ebasco.stats.calib = ds.waveforms[curr_sta.replace('.','_')][tag_comp][0].stats.calib
    
    trc_x_vel_ebasco.stats.network = ds.waveforms[curr_sta.replace('.','_')][tag_comp][0].stats.network
    trc_x_vel_ebasco.stats.station = ds.waveforms[curr_sta.replace('.','_')][tag_comp][0].stats.station
    trc_x_vel_ebasco.stats.location = ds.waveforms[curr_sta.replace('.','_')][tag_comp][0].stats.location
    trc_x_vel_ebasco.stats.channel = ds.waveforms[curr_sta.replace('.','_')][tag_comp][0].stats.channel
    trc_x_vel_ebasco.stats.starttime = starttime
    trc_x_vel_ebasco.stats.sampling_rate = ds.waveforms[curr_sta.replace('.','_')][tag_comp][0].stats.sampling_rate
    trc_x_vel_ebasco.stats.delta = ds.waveforms[curr_sta.replace('.','_')][tag_comp][0].stats.delta
    trc_x_vel_ebasco.stats.npts = len(VEL_X_eb)
    trc_x_vel_ebasco.stats.calib = ds.waveforms[curr_sta.replace('.','_')][tag_comp][0].stats.calib   
    
    trc_x_dis_ebasco.stats.network = ds.waveforms[curr_sta.replace('.','_')][tag_comp][0].stats.network
    trc_x_dis_ebasco.stats.station = ds.waveforms[curr_sta.replace('.','_')][tag_comp][0].stats.station
    trc_x_dis_ebasco.stats.location = ds.waveforms[curr_sta.replace('.','_')][tag_comp][0].stats.location
    trc_x_dis_ebasco.stats.channel = ds.waveforms[curr_sta.replace('.','_')][tag_comp][0].stats.channel
    trc_x_dis_ebasco.stats.starttime = starttime
    trc_x_dis_ebasco.stats.sampling_rate = ds.waveforms[curr_sta.replace('.','_')][tag_comp][0].stats.sampling_rate
    trc_x_dis_ebasco.stats.delta = ds.waveforms[curr_sta.replace('.','_')][tag_comp][0].stats.delta
    trc_x_dis_ebasco.stats.npts = len(DIS_X_eb)
    trc_x_dis_ebasco.stats.calib = ds.waveforms[curr_sta.replace('.','_')][tag_comp][0].stats.calib       
        
    trc_x_dis_ebasco.stats['permanent_displacement'] = DIS_X_eb[-1]
    
    return trc_x_acc_ebasco, trc_x_vel_ebasco, trc_x_dis_ebasco 
    
# add spectra (2 array of data -> obspy objects)
def add_spectra(st,pers,vals):
    trace1 = trace.Trace(data=array([]), header=st.stats); 
    trace2 = trace.Trace(data=array([]), header=st.stats); 
    st = Stream(traces=[trace1, trace2])
    st[0].data = pers
    st[1].data = vals
    return st  

def valid_asdf(ifile):
    err_str0 = "input file '" + ifile + "' does not exist!"
    if not os_path_isfile(ifile): raise argparse.ArgumentTypeError(err_str0)
    
    err_str2 = "input file '" + ifile + "' is not an ASDF volume"
    try:
        ds = pyasdf.ASDFDataSet(ifile, mode='r')
        return [ifile, ds]
    except: raise argparse.ArgumentTypeError(err_str2)


def main():
    
    if len(sys_argv) < 2: sys_argv.append('--help')
    
    ret_code = eBASCO(' '.join(sys_argv[1:]))
    if ret_code['exit_status'] == 0 and ret_code['out_string']: print ret_code['out_string']
    if ret_code['exit_message']: print ret_code['exit_message']
    sys_exit(ret_code['exit_status'])

#-------------------------------------------------------------

##-------------------------------------------------------------

if __name__ == "__main__": main()
